#!/usr/bin/env python

## Twisted Life: A game that uses a variant John H. Conway's Life cellular
## automaton in which Life forms twist around the center of the universe.
## Copyright (C) 2007  Paul Reiners
##
## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module is the GUI for Twisted Life.
"""


#Import Modules
import os, pygame, random
from pygame.locals import *
import twisted_life
import data
from buttons import CommandButton

if not pygame.font: print 'Warning, fonts disabled'
if not pygame.mixer: print 'Warning, sound disabled'

#game constants
NUM_LIVES         = 4
HEIGHT = WIDTH    = 468
SCREENRECT        = Rect(0, 0, HEIGHT, WIDTH)
MAX_ROUND_SCORE   = 1000
SCORE_REDUCER     = 0.99
UNIVERSE_RADIUS   = 32
BOTTOM_PANEL_Y    = 460

# globals
score                   = 0
fractionOfMaxRoundScore = 1.0
lose_sound              = None
win_sound               = None
topScore                = 0
livesLeft               = 4

class dummysound:
    def play(self): pass

def load_sound(file):
    if not pygame.mixer: return dummysound()
    file = os.path.join('data', file)
    try:
        sound = pygame.mixer.Sound(file)
        return sound
    except pygame.error:
        print 'Warning, unable to load,', file
    return dummysound()

def initializePlayer(universe):
    onLiveSquare = True
    universeDiameter = universe.diameter()
    while onLiveSquare:
        playerX = random.randint(0, universeDiameter - 1)
        playerY = random.randint(0, universeDiameter - 1)
        onLiveSquare = universe.isAlive(playerX, playerY)

    return [playerX, playerY]
    
class Score(pygame.sprite.Sprite):
    def __init__(self):
        pygame.sprite.Sprite.__init__(self)
        self.font = pygame.font.Font(None, 20)
        self.color = Color('white')
        self.update()
        self.rect = self.image.get_rect().move(10, BOTTOM_PANEL_Y)

    def update(self):
        global screen
        global livesLeft
        msg = ("Score: %d  Round Score: %d  Lives: %d  High Score: %d    " %
                       (score, int(round(fractionOfMaxRoundScore * MAX_ROUND_SCORE)), livesLeft, topScore))
        self.image = self.font.render(msg, 1, self.color, Color('black'))
    
class NewGameButton(pygame.sprite.Sprite):
    def __init__(self, rect):
        pygame.sprite.Sprite.__init__(self)
        self.font = pygame.font.Font(None, 20)
        self.color = Color('white')
        self.update()
        self.rect = rect

    def update(self):
        msg = ("New Game")
        self.image = self.font.render(msg, 1, self.color, Color('gray'))

def updateTextSprites(all, screen, background):
    # clear/erase the last drawn sprites
    all.clear(screen, background)

    #update all the sprites
    all.update()

    #draw the scene
    dirty = all.draw(screen)
    pygame.display.update(dirty)

class NewGameCommand:
    def __init__(self, app):
        self.app = app
    def do(self):
        self.app.startNewGame()

class TwistedLifeApp:
    def __init__(self):
        global lose_sound
        global win_sound
    #Initialize Everything
        pygame.init()
        self.screen = pygame.display.set_mode((WIDTH, HEIGHT + 15))
        self.background = pygame.Surface(SCREENRECT.size)
        pygame.display.set_caption('Twisted Life')
        pygame.mouse.set_visible(1)

        #load the sound effects and music
        lose_sound = load_sound('boom.wav')
        win_sound = load_sound('secosmic_lo.wav')
        self.monk_music = load_sound('rule225batman.wav')

        self.buttons = []

        # Initialize Game Groups
        self.all = pygame.sprite.RenderUpdates()

        #assign default groups to each sprite class
        Score.containers = self.all

        universeDiameter = 2 * UNIVERSE_RADIUS + 1
        
        self.rects = []
        cellWidth = WIDTH / universeDiameter
        cellHeight = HEIGHT / universeDiameter
        for i in range(universeDiameter):
            row = []
            y = i * cellHeight
            for j in range(universeDiameter):
                x = j * cellWidth
                rect = pygame.Rect(x, y, cellWidth, cellHeight)
                row.append(rect)
            self.rects.append(row)

        newGameRect = Rect(352, BOTTOM_PANEL_Y, 100, 45)
        self.addButton(CommandButton(newGameRect, NewGameCommand(self)))

        #initialize our starting sprites
        if pygame.font:
            self.all.add(Score())
            self.all.add(NewGameButton(newGameRect))

    def checkResult(self, playerX, playerY, universe):
        playerX = (playerX + universe.diameter()) % universe.diameter()
        playerY = (playerY + universe.diameter()) % universe.diameter()
        playerCoords = (playerX, playerY)
        self.display(universe, playerCoords)
        if universe.isAlive(playerX, playerY):
            lose_sound.play()

            return 'lost'
        elif playerCoords == universe.getCenter():
            win_sound.play()

            return 'won'
        else:
            return ""

    def display(self, universe, playerCoords):
        for i in range(universe.diameter()):
            for j in range(universe.diameter()):
                rect = self.rects[j][i]
                if universe.isAlive(j, i):
                    color = Color('black')
                else:
                    color = Color('white')
                pygame.draw.rect(self.screen, color, rect)
        centerRect = self.rects[universe.radius][universe.radius]
        if universe.isAlive(universe.radius, universe.radius):
            color = Color('red')
        else:
            color = Color('green')
        pygame.draw.rect(self.screen, color, centerRect)

        updateTextSprites(self.all, self.screen, self.background)

        playerRect = self.rects[playerCoords[0]][playerCoords[1]]
        if universe.isAlive(playerCoords[0], playerCoords[1]):
            color = Color('red')
        else:
            color = Color('blue')
        pygame.draw.rect(self.screen, color, playerRect)
        #Draw Everything
        pygame.display.flip()

    def roundLoop(self, playerCoords, universe):
        global fractionOfMaxRoundScore
        global lose_sound
        global win_sound

        fractionOfMaxRoundScore = 1.0
        
        playerX = playerCoords[0]
        playerY = playerCoords[1]
        #Main Loop
        while 1:
            self.clock.tick(60)

        #Handle Input Events
            moved = False
            for event in pygame.event.get():
                if event.type == QUIT:
                    return 'quit'
                elif (event.type == KEYDOWN):
                    if (event.key == K_ESCAPE):
                        return 'quit'
                    elif (event.key == K_UP):
                        playerX -= 1
                        moved = True
                    elif (event.key == K_DOWN):
                        playerX += 1
                        moved = True
                    elif (event.key == K_LEFT):
                        playerY -= 1
                        moved = True
                    elif (event.key == K_RIGHT):
                        playerY += 1
                        moved = True
                    if moved:
                        result = self.checkResult(playerX, playerY, universe)
                        if (result != ""):
                            return result

            if not moved:                    
                pressedKeys = pygame.key.get_pressed()
                if (pressedKeys[K_UP]):
                    playerX -= 1
                    moved = True
                elif (pressedKeys[K_DOWN]):
                    playerX += 1
                    moved = True
                elif (pressedKeys[K_LEFT]):
                    playerY -= 1
                    moved = True
                elif (pressedKeys[K_RIGHT]):
                    playerY += 1
                    moved = True
                if moved:
                    result = self.checkResult(playerX, playerY, universe)
                    if (result != ""):
                        return result

            universe.nextGen()
            playerX = (playerX + universe.diameter()) % universe.diameter()
            playerY = (playerY + universe.diameter()) % universe.diameter()
            playerCoords = (playerX, playerY)
            self.display(universe, playerCoords)
            if universe.isAlive(playerX, playerY):
                lose_sound.play()
                return 'lost'
            elif playerCoords == universe.getCenter():
                win_sound.play()
                return 'won'
            fractionOfMaxRoundScore *= SCORE_REDUCER
    
    def addButton(self, button):
        self.buttons = self.buttons + [button]

    def startNewGame(self):
        print "Starting new game"
        self.newGame = True

    def loopForNewGame(self):
        self.newGame = False
        exit = False
        while not exit:
            exit = self.handleEvents()
            # self.drawButtons()
            updateTextSprites(self.all, self.screen, self.background)
            pygame.display.update()

    def handleEvents(self):
            exit = False
            for event in pygame.event.get():
                if event.type == QUIT:
                    exit = True
                elif event.type == KEYDOWN:
                    if event.key == K_ESCAPE:
                        exit = True
                elif event.type == MOUSEBUTTONDOWN:
                    self.handleMouseDown(pygame.mouse.get_pos())
                    if self.newGame:
                        exit = True
            return exit

    def handleMouseDown(self, (x, y)):
        for button in self.buttons:
            button.handleMouseDown(x, y)
 
    def playGame(self):
        global score
        global livesLeft
        global topScore

        topScore = int(data.load('high_scores.txt').read())

        universe = twisted_life.TwistedLife(UNIVERSE_RADIUS)

        universe.nextGen()

    #Display The Background
        pygame.display.flip()
        
    #Prepare Game Objects
        self.clock = pygame.time.Clock()
        lossCount = 0
        roundNum = 1
        score = 0

        updateTextSprites(self.all, self.screen, self.background)

        while lossCount < NUM_LIVES:
            livesLeft = NUM_LIVES - lossCount
            print 'Round ' + str(roundNum) + ': '
            playerCoords = initializePlayer(universe)
            self.display(universe, playerCoords)

            playerReady = False
            pygame.event.clear()
            while not playerReady:
                self.clock.tick(60)

            # Wait for player to start round.                
                for event in pygame.event.get():
                    if event.type == QUIT:
                        return
                    elif (event.type == KEYDOWN):
                        if (event.key == K_ESCAPE):
                            return
                        elif (event.key == K_UP):
                            playerCoords[0] = (playerCoords[0] - 1 + universe.diameter()) % universe.diameter()
                            playerReady = True
                            
                            break
                        elif (event.key == K_DOWN):
                            playerCoords[0] = (playerCoords[0] + 1 + universe.diameter()) % universe.diameter()
                            playerReady = True

                            break
                        elif (event.key == K_LEFT):
                            playerCoords[1] = (playerCoords[1] - 1 + universe.diameter()) % universe.diameter()
                            playerReady = True

                            break
                        elif (event.key == K_RIGHT):
                            playerCoords[1] = (playerCoords[1] + 1 + universe.diameter()) % universe.diameter()
                            playerReady = True

                            break
                        
            result = self.roundLoop(playerCoords, universe)
            if result == 'lost':
                print '\tYou ran into a live cell.  You lose the round.'
                lossCount += 1
            elif result == 'won':
                print '\tYou reached the goal.  You win the round.'
                score += int(round(fractionOfMaxRoundScore * MAX_ROUND_SCORE))
            elif result == 'quit':
                return

            print "Your score:", score

            updateTextSprites(self.all, self.screen, self.background)

            pygame.time.wait(2000)
            universe = twisted_life.TwistedLife(UNIVERSE_RADIUS)
            universe.nextGen()
            roundNum += 1
        
    #Game Over
        livesLeft = 0
        if score > topScore:
            scoreFile = data.load('high_scores.txt', 'w')
            scoreFile.write(str(score))
            scoreFile.close()            

        updateTextSprites(self.all, self.screen, self.background)

        print "Your final score:", score
        self.monk_music.play()
 
    def drawButtons(self):
        for button in self.buttons:
            button.draw(self.screen)

    def run(self):
        """this function is called when the program starts.
           it runs in a loop until the function returns."""

        self.playGame()
        updateTextSprites(self.all, self.screen, self.background)

        # Wait for new game or quit.
        while True:
            self.loopForNewGame()
            if self.newGame:
                self.playGame()
            else:
                break
        
        pygame.quit()

def main():
    TwistedLifeApp().run()

#this calls the 'main' function when this script is executed
if __name__ == '__main__': main()
